@@ -15,7 +15,7 @@ import sys
 import getopt
 
 SOURCE = 'Armooo-PrintProxy-1'
-PRINT_CLOUD_SERICE_ID = 'cloudprint'
+PRINT_CLOUD_SERVICE_ID = 'cloudprint'
 CLIENT_LOGIN_URL = '/accounts/ClientLogin'
 PRINT_CLOUD_URL = '/cloudprint/'
 
@@ -28,6 +28,9 @@ class CloudPrintProxy(object):
         self.cups= cups.Connection()
         self.proxy =  platform.node() + '-Armooo-PrintProxy'
         self.auth_path = os.path.expanduser('~/.cloudprintauth')
+        self.xmpp_auth_path = os.path.expanduser('~/.cloudprintsaslauth')
+        self.username = None
+        self.password = None
 
     def get_auth(self):
         if self.auth:
@@ -37,20 +40,43 @@ class CloudPrintProxy(object):
             if auth:
                 return auth
 
-            username = raw_input('Google username: ')
-            password = getpass.getpass()
-
             r = rest.REST('https://www.google.com', debug=False)
-            auth_response = r.post(
-                CLIENT_LOGIN_URL,
-                {
-                    'accountType': 'GOOGLE',
-                    'Email': username,
-                    'Passwd': password,
-                    'service': PRINT_CLOUD_SERICE_ID,
-                    'source': SOURCE,
-                },
-                'application/x-www-form-urlencoded')
+            try:
+                auth_response = r.post(
+                    CLIENT_LOGIN_URL,
+                    {
+                        'accountType': 'GOOGLE',
+                        'Email': self.username,
+                        'Passwd': self.password,
+                        'service': PRINT_CLOUD_SERVICE_ID,
+                        'source': SOURCE,
+                    },
+                    'application/x-www-form-urlencoded')
+                xmpp_response = r.post(CLIENT_LOGIN_URL,
+                    {
+                        'accountType': 'GOOGLE',
+                        'Email': self.username,
+                        'Passwd': self.password,
+                        'service': 'mail',
+                        'source': SOURCE,
+                    },
+                    'application/x-www-form-urlencoded')
+                jid = self.username if '@' in self.username else self.username + '@gmail.com'
+                sasl_token = ('\0%s\0%s' % (jid, xmpp_response['Auth'])).encode('base64')
+                file(self.xmpp_auth_path, 'w').write(sasl_token)
+            except rest.REST.RESTException, e:
+                if 'InvalidSecondFactor' in e.msg:
+                    raise rest.REST.RESTException(
+                        '2-Step',
+                        '403',
+                        'You have 2-Step authentication enabled on your '
+                        'account. \n\nPlease visit '
+                        'https://www.google.com/accounts/IssuedAuthSubTokens '
+                        'to generate an application-specific password.'
+                    )
+                else:
+                    raise
+
             self.set_auth(auth_response['Auth'])
             return self.auth
 
@@ -185,7 +211,7 @@ class CloudPrintProxy(object):
     def finish_job(self, job_id):
         r = self.get_rest()
         r.post(
-            PRINT_CLOUD_URL + '/control',
+            PRINT_CLOUD_URL + 'control',
             {
                 'output' : 'json',
                 'jobid': job_id,
@@ -198,7 +224,7 @@ class CloudPrintProxy(object):
     def fail_job(self, job_id):
         r = self.get_rest()
         r.post(
-            PRINT_CLOUD_URL + '/control',
+            PRINT_CLOUD_URL + 'control',
             {
                 'output' : 'json',
                 'jobid': job_id,
@@ -223,6 +249,20 @@ class PrinterProxy(object):
     def delete(self):
         return self.cpp.delete_printer(self.id)
 
+class App(object):
+    def __init__(self, cups_connection=None, cpp=None, printers=None, pidfile_path=None):
+        self.cups_connection = cups_connection
+        self.cpp = cpp
+        self.printers = printers
+        self.pidfile_path = pidfile_path
+        self.stdin_path = '/dev/null'
+        self.stdout_path = '/dev/tty'
+        self.stderr_path = '/dev/tty'
+        self.pidfile_timeout = 5
+
+    def run(self):
+        process_jobs(self.cups_connection, self.cpp, self.printers)
+
 
 def sync_printers(cups_connection, cpp):
     local_printer_names = set(cups_connection.getPrinters().keys())
@@ -278,38 +318,73 @@ def process_job(cups_connection, cpp, printer, job):
             'Authorization' : 'GoogleLogin auth=%s' % cpp.get_auth()
         })
         options = json.loads(urllib2.urlopen(request).read())
-        del options['request']
+        if 'request' in options: del options['request']
         options = dict( (str(k), str(v)) for k, v in options.items() )
 
         cpp.finish_job(job['id'])
 
         cups_connection.printFile(printer.name, tmp.name, job['title'], options)
         os.unlink(tmp.name)
-        print "SUCCESS ",job['title']
+        print "SUCCESS ",job['title'].encode('unicode-escape')
 
     except:
         cpp.fail_job(job['id'])
-        print "ERROR",job['title']
+        print "ERROR",job['title'].encode('unicode-escape')
 
 def process_jobs(cups_connection, cpp, printers):
     while True:
-        for printer in printers:
-            for job in printer.get_jobs():
-                process_job(cups_connection, cpp, printer, job)
-        time.sleep(60)
+        try:
+            for printer in printers:
+                for job in printer.get_jobs():
+                    process_job(cups_connection, cpp, printer, job)
+            wait_for_new_job(file(cpp.xmpp_auth_path).read())
+        except Exception, e:
+            print e
+            print "ERROR: Couldn't Connect to Cloud Service. Will Try again in 60 Seconds";
+            time.sleep(60)
+
+def wait_for_new_job(sasl_token):
+    # https://developers.google.com/cloud-print/docs/rawxmpp
+    import ssl, socket
+    from xml.etree.ElementTree import iterparse, tostring
+    xmpp = ssl.wrap_socket(socket.socket())
+    xmpp.connect(('talk.google.com', 5223))
+    parser = iterparse(xmpp, ('start', 'end'))
+    def msg(msg=' '):
+        xmpp.write(msg)
+        stack = 0
+        for event, el in parser:
+            if event == 'start' and el.tag.endswith('stream'):
+                continue
+            stack += 1 if event == 'start' else -1
+            if stack == 0:
+                assert not el.tag.endswith('failure') and not el.tag.endswith('error') and not el.get('type') == 'error', tostring(el)
+                return el
+    msg('<stream to="gmail.com" version="1.0" xmlns="http://etherx.jabber.org/streams">')
+    msg('<auth xmlns="urn:ietf:params:xml:ns:xmpp-sasl" mechanism="X-GOOGLE-TOKEN">%s</auth>' % sasl_token)
+    msg('<s:stream to="gmail.com" version="1.0" xmlns:s="http://etherx.jabber.org/streams" xmlns="jabber:client">')
+    iq = msg('<iq type="set"><bind xmlns="urn:ietf:params:xml:ns:xmpp-bind"><resource>Armooo</resource></bind></iq>')
+    bare_jid = iq[0][0].text.split('/')[0]
+    msg('<iq type="set" to="%s"><subscribe xmlns="google:push"><item channel="cloudprint.google.com" from="cloudprint.google.com"/></subscribe></iq>' % bare_jid)
+    return msg()
 
 def usage():
-    print sys.argv[0] + ' [-d][-l][-h] [-p pid_file]'
+    print sys.argv[0] + ' [-d][-l][-h] [-p pid_file] [-a account_file]'
     print '-d\t\t: enable daemon mode (requires the daemon module)'
     print '-l\t\t: logout of the google account'
     print '-p pid_file\t: path to write the pid to (default cloudprint.pid)'
+    print '-a account_file\t: path to google account ident data (default ~/.cloudprintauth)'
+    print '\t\t account_file format:\t <Google username>'
+    print '\t\t\t\t\t <Google password>'
     print '-h\t\t: display this help'
 
 def main():
-    opts, args = getopt.getopt(sys.argv[1:], 'dlhp:')
+    opts, args = getopt.getopt(sys.argv[1:], 'dlhp:a:')
     daemon = False
     logout = False
     pidfile = None
+    authfile = None
+    saslauthfile = None
     for o, a in opts:
         if o == '-d':
             daemon = True
@@ -317,6 +392,9 @@ def main():
             logout = True
         elif o == '-p':
             pidfile = a
+        elif o == '-a':
+            authfile = a
+            saslauthfile = authfile+".sasl"
         elif o =='-h':
             usage()
             sys.exit()
@@ -325,12 +403,31 @@ def main():
 
     cups_connection = cups.Connection()
     cpp = CloudPrintProxy()
+    if authfile:
+        cpp.auth_path = authfile
+        cpp.xmpp_auth_path = saslauthfile
 
     if logout:
         cpp.del_saved_auth()
         print 'logged out'
         return
 
+    # Check if authentification is needed
+    if not cpp.get_saved_auth():
+        if authfile:
+            if os.path.exists(authfile):
+                account_file = open(authfile)
+                cpp.username = account_file.next().rstrip()
+                cpp.password = account_file.next().rstrip()
+                account_file.close()
+
+            else:
+                print 'Unable to find account file'
+                return
+        else:
+          cpp.username = raw_input('Google username: ')
+          cpp.password = getpass.getpass()
+
     #try to login
     while True:
         try:
@@ -350,15 +447,20 @@ def main():
 
     if daemon:
         try:
-            import daemon
+            from daemon import runner
         except ImportError:
             print 'daemon module required for -d'
-            print '\tpip install daemon'
+            print '\tyum install python-daemon, or apt-get install python-daemon, or pip install python-daemon'
             sys.exit(1)
-        daemon.daemonize(pidfile)
-
-    process_jobs(cups_connection, cpp, printers)
-
+        
+        app = App(cups_connection=cups_connection,
+                  cpp=cpp, printers=printers,
+                  pidfile_path=os.path.abspath(pidfile))
+        sys.argv=[sys.argv[0], 'start']
+        daemon_runner = runner.DaemonRunner(app)
+        daemon_runner.do_action()
+    else:
+        process_jobs(cups_connection, cpp, printers)
 
 if __name__ == '__main__':
     main()
